#ifndef lint
static char sccs_id[] = "@(#) xtiff.c 1.5@(#)";
#endif /*lint */

/***
  xtiff.c - bashed-about version of xtiff for CERN scanned preprints.
	    this is a bit slow & stupid about rotations & zooms - the actual 
	    image is rotated & subsampled (for zooms) and another buffer
	    big enough to hold the image is used.  things can probably be
	    speeded up using cut buffers ...
	    although the stuff for other than 1-bit images has been left in
	    and code's been written to rotate & zoom the other formats, it's
	    not been tested & isn't useful for scanned preprints & so may be
	    removed later on.
	    no warranty, etc etc etc.  you know all that.  the stuff from the
	    previous authors is below.

  dave robinson (dave@preprints.cern.ch until august 1995)

***/
/*
 * xtiff - view a TIFF file in an X window
 *
 * Dan Sears
 * Chris Sears
 *
 * Copyright 1991 by Digital Equipment Corporation, Maynard, Massachusetts.
 *
 *                      All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of Digital not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
 * DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * Revision 1.0  90/05/07
 *      Initial release.
 * Revision 2.0  90/12/20
 *      Converted to use the Athena Widgets and the Xt Intrinsics.
 *
 * Notes:
 *
 * According to the TIFF 5.0 Specification, it is possible to have
 * both a TIFFTAG_COLORMAP and a TIFFTAG_COLORRESPONSECURVE.  This
 * doesn't make sense since a TIFFTAG_COLORMAP is 16 bits wide and
 * a TIFFTAG_COLORRESPONSECURVE is tfBitsPerSample bits wide for each
 * channel.  This is probably a bug in the specification.
 * In this case, TIFFTAG_COLORRESPONSECURVE is ignored.
 * This might make sense if TIFFTAG_COLORMAP was 8 bits wide.
 *
 * TIFFTAG_COLORMAP is often incorrectly written as ranging from
 * 0 to 255 rather than from 0 to 65535.  CheckAndCorrectColormap()
 * takes care of this.
 *
 * Only ORIENTATION_TOPLEFT is supported correctly.  This is the
 * default TIFF and X orientation.  Other orientations will be
 * displayed incorrectly.
 *
 * There is no support for or use of 3/3/2 DirectColor visuals.
 * TIFFTAG_MINSAMPLEVALUE and TIFFTAG_MAXSAMPLEVALUE are not supported.
 *
 * Only TIFFTAG_BITSPERSAMPLE values that are 1, 2, 4 or 8 are supported.
 */
#include <math.h>
#include <stdio.h>

#include <tiffio.h>
/*
  version 1.14 of libtiff does not define TIFFTAG_COLORRESPONSECURVE
*/
#ifndef TIFFTAG_COLORRESPONSECURVE
#define TIFFTAG_COLORRESPONSECURVE      301     /* RGB response curve */
#endif	/* TIFFTAG_COLORRESPONSECURVE */

#include <X11/Xatom.h>

/* 
  emergency measure, to stop (some) confilcts w/gcc 
*/
/*#define _GCC_WCHAR_T*/	
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xproto.h>
#include <X11/Shell.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Dialog.h>
#include <X11/cursorfont.h>
#define XK_MISCELLANY
#include <X11/keysymdef.h>
#include "xtifficon.h"
#include "buttons.h"

#define TIFF_GAMMA      "2.2"     /* default gamma from the TIFF 5.0 spec */
#define ROUND(x)        (u_short) ((x) + 0.5)
#define SCALE(x, s)     (((x) * 65535L) / (s))
#define MCHECK(m)       if (!m) { \
				(void)fprintf(stderr, "malloc failed\n");\
				 exit(0); \
			}
#define MIN(a, b)       (((a) < (b)) ? (a) : (b))
#define MAX(a, b)       (((a) > (b)) ? (a) : (b))
#define VIEWPORT_WIDTH  700
#define VIEWPORT_HEIGHT 500
#define KEY_TRANSLATE   20
#define ZOOMFACTOR 	1

#ifdef __STDC__
#define PP(args)    args
#else
#define PP(args)    ()
#endif

void main PP((int argc, char **argv));
void OpenTIFFFile PP((void));
void CountTIFFPages PP((void));
void GetTIFFHeader PP((void));
void SetNameLabel PP((void));
void CheckAndCorrectColormap PP((void));
void SimpleGammaCorrection PP((void));
void GetVisual PP((void));
Boolean SearchVisualList PP((int image_depth,
    int visual_class, Visual **visual));
u_long allocimage PP((void));
void GetTIFFImage PP((void));
void CreateXImage PP((int state));
XtCallbackProc SelectProc PP((Widget w, caddr_t unused_1, caddr_t unused_2));
void QuitProc PP((void));
void NextProc PP((void));
void PreviousProc PP((void));
void GrowProc PP((void));
void ShrinkProc PP((void));
void LargestProc PP((void));
void BestFitProc PP((void));
void ZeroProc PP((void));
void NinetyProc PP((void));
void OneEightyProc PP((void));
void TwoSeventyProc PP((void));
void PageProc PP((int direction));
void EventProc PP((Widget widget, caddr_t unused, XEvent *event));
void FirstPageProc PP((void));
void LastPageProc PP((void));
void GotoPageProc PP((void));
int XTiffErrorHandler PP((Display *display, XErrorEvent *error_event));
void zoomproc PP((int flag, int zoomordelta));
void rotationproc PP((int rotationfactor));
void Usage PP((void));
void busy PP((Boolean toggle));

void GetBestFit PP((void));
void redisplay PP((int status));


/*
  these area all in transform.c
*/
extern int 	transformimage PP((int status));
extern void 	setzoomfactor PP((int newzoomfactor));
extern int 	getzoomfactor PP((void));
extern int 	getotationfacor PP((void));
extern void 	setrotationfactor PP((int newrotationfactor));
extern u_long	getwidth PP((void));
extern u_long	getheight PP((void));
extern u_long	getrealwidth PP((void));
extern u_long	getrealheight PP((void));
extern void 	seteffectivewidth PP((u_long length));
extern void 	seteffectiveheight PP((u_long length));

/*
  these are all in help.c
*/
extern Boolean amhelpful PP((void));
extern void behelpful PP((void));

int xtVersion = XtSpecificationRelease;     /* xtiff depends on R4 or higher */

/*
  zoomflags
*/
enum { ZOOMDELTA = 0 , ZOOMSET = 1 };

/*
  status flags for CreateXIamge
*/
enum { INITSTATE = 0, PAGECHANGE = 1, ZOOMCHANGE = 2, ROTATECHANGE = 3 };

/*
 * Xt data structures
 */
Widget  shellWidget, 
	popupWidget,
	dialogueWidget,
	formWidget, 
	labelWidget, 
	boxWidget,
	imageWidget;

enum { 	ButtonQuit = 0,
	ButtonPreviousPage = 1, 
	ButtonNextPage = 2,
	ButtonFirstPage = 3,
	ButtonLastPage = 4,
	ButtonShrink = 5, 
       	ButtonGrow= 6, 
	ButtonLargest = 7,
       	ButtonBestFit = 8, 
	Button0 = 9, 
	Button90 = 10, 
	Button180 = 11,
	Button270 = 12,
	ButtonGotoDialoguePopup = 13,
	ButtonHelp = 14,

	/* 
	  the following are psuedo-buttons returned from the select page 
	  dialogue widget
	*/
	ButtonGotoPage = 1015,
	ButtonDontGotoPage = 1016 };

static XrmOptionDescRec shellOptions[] = {
    { "-help", "*help", XrmoptionNoArg, (caddr_t) "True" },
    { "-gamma", "*gamma", XrmoptionSepArg, NULL },
    { "-usePixmap", "*usePixmap", XrmoptionSepArg, NULL },
    { "-viewportWidth", "*viewportWidth", XrmoptionSepArg, NULL },
    { "-viewportHeight", "*viewportHeight", XrmoptionSepArg, NULL },
    { "-translate", "*translate", XrmoptionSepArg, NULL },
    { "-verbose", "*verbose", XrmoptionSepArg, NULL },
    { "-zf", "*zf", XrmoptionSepArg, NULL }
};

typedef struct {
    Boolean help;
    float gamma;
    Boolean usePixmap;
    int viewportWidth;
    int viewportHeight;
    int translate;
    Boolean verbose;
    int	zf;
} AppData, *AppDataPtr;

AppData appData;

XtResource clientResources[] = {
    {
        "help", XtCBoolean, XtRBoolean, sizeof(Boolean),
        XtOffset(AppDataPtr, help), XtRImmediate, (XtPointer) False
    }, {
        "gamma", "Gamma", XtRFloat, sizeof(float),
        XtOffset(AppDataPtr, gamma), XtRString, (XtPointer) TIFF_GAMMA
    }, {
        "usePixmap", "UsePixmap", XtRBoolean, sizeof(Boolean),
        XtOffset(AppDataPtr, usePixmap), XtRImmediate, (XtPointer) True
    }, {
        "viewportWidth", "ViewportWidth", XtRInt, sizeof(int),
        XtOffset(AppDataPtr, viewportWidth), XtRImmediate,
        (XtPointer) VIEWPORT_WIDTH
    }, {
        "viewportHeight", "ViewportHeight", XtRInt, sizeof(int),
        XtOffset(AppDataPtr, viewportHeight), XtRImmediate,
        (XtPointer) VIEWPORT_HEIGHT
    }, {
        "translate", "Translate", XtRInt, sizeof(int),
        XtOffset(AppDataPtr, translate), XtRImmediate, (XtPointer) KEY_TRANSLATE
    }, {
        "verbose", "Verbose", XtRBoolean, sizeof(Boolean),
        XtOffset(AppDataPtr, verbose), XtRImmediate, (XtPointer) True
    }, {
	"zf", "Zf", XtRInt, sizeof(int), XtOffset(AppDataPtr, zf),
	XtRImmediate, (XtPointer) ZOOMFACTOR
    }
};

Arg formArgs[] = {
    { XtNresizable, True }
};

Arg popupArgs[] = {
    { XtNwidth, 300 },
    { XtNheight, 150 }
};

Arg dialogueArgs[] = {
    { XtNlabel, (long) "Type the page number, click `goto' or `cancel'" },
    { XtNvalue, (long) "1" },
    { XtNborderWidth, 0 }
};

Arg boxArgs[] = {
    { XtNorientation, (long) XtEhorizontal },
    { XtNtop, XtChainTop },
    { XtNleft, XtChainLeft },
    { XtNbottom, XtChainTop },
    { XtNright, XtChainLeft }
};


Arg labelArgs[] = {
    { XtNresizable, False },
    { XtNwidth, 400 },
    { XtNborderWidth, 0 },
    { XtNjustify, XtJustifyLeft },
    { XtNtop, XtChainTop },
    { XtNleft, XtChainLeft },
    { XtNbottom, XtChainTop },
    { XtNright, XtChainLeft }
};

Arg imageArgs[] = {
    { XtNresizable, True },
    { XtNborderWidth, 0 },
    { XtNtop, XtChainTop },
    { XtNleft, XtChainLeft },
    { XtNbottom, XtChainTop },
    { XtNright, XtChainLeft }
};


/*
 * X data structures
 */
Colormap            xColormap;
Display *           xDisplay;
Pixmap              xImagePixmap;
Visual *            xVisual;
XImage *            xImage;
GC                  xWinGc;
int                 xImageDepth, xScreen, xRedMask, xGreenMask, xBlueMask,
                    xOffset = 0, yOffset = 0, grabX = -1, grabY = -1;
u_char              basePixel = 0;

/*
 * TIFF data structures
 */
TIFF *              tfFile = NULL;
u_long              tfImageWidth, tfImageHeight;
u_short             tfBitsPerSample, tfSamplesPerPixel, tfPlanarConfiguration,
                    tfPhotometricInterpretation, tfGrayResponseUnit,
                    tfImageDepth, tfBytesPerRow;
int                 tfDirectory = 0,  tfNumPages = 1;
double              tfUnitMap, tfGrayResponseUnitMap[] = {
                        -1, -10, -100, -1000, -10000, -100000
                    };

/*
 * display data structures
 */
double              *dRed, *dGreen, *dBlue;

/*
 * shared data structures
 */
u_short *           redMap = NULL, *greenMap = NULL, *blueMap = NULL,
                    *grayMap = NULL, colormapSize;
u_char *            imageMemory = NULL, *zoomMemory = NULL;
char *              fileName;

int		    isize;
int		    imagebitsize;

Pixmap		    buttonpixmaps[15];

void
main(argc, argv)
    int argc;
    char ** argv;
{
    int		w;
    Widget widget_list[15];
    Arg args[5];

    static char	*button_bitmaps[] = {   quit_bits,
					prev_bits,
					next_bits,
					first_bits,
					last_bits,
					shrink_bits,
					grow_bits,
					full_bits,
					fit_bits,
					zero_bits,
					ninety_bits,
					oneeighty_bits,
					twoseventy_bits,
					goto_bits,
					help_bits
				     };

    setbuf(stdout, NULL); setbuf(stderr, NULL);

    shellWidget = XtInitialize(argv[0], "XTiff", shellOptions,
        XtNumber(shellOptions), &argc, argv);

    XSetErrorHandler(XTiffErrorHandler);

    XtGetApplicationResources(shellWidget, &appData,
        (XtResourceList) clientResources, (Cardinal) XtNumber(clientResources),
        (ArgList) NULL, (Cardinal) 0);

    if ((argc <= 1) || (argc > 2) || appData.help)
        Usage();

    if (appData.verbose == False) {
        TIFFSetErrorHandler(0);
        TIFFSetWarningHandler(0);
    }

    fileName = argv[1];

    xDisplay = XtDisplay(shellWidget);
    xScreen = DefaultScreen(xDisplay);

    OpenTIFFFile();
    CountTIFFPages();
    GetTIFFHeader();
    SimpleGammaCorrection();
    GetVisual();
    GetTIFFImage();

    /*
     * Send visual, colormap, depth and iconPixmap to shellWidget.
     * Sending the visual to the shell is only possible with the advent of R4.
     */
    XtSetArg(args[0], XtNvisual, xVisual);
    XtSetArg(args[1], XtNcolormap, xColormap);
    XtSetArg(args[2], XtNdepth,
        xImageDepth == 1 ? DefaultDepth(xDisplay, xScreen) : xImageDepth);
    XtSetArg(args[3], XtNiconPixmap,
        XCreateBitmapFromData(xDisplay, RootWindow(xDisplay, xScreen),
            xtifficon_bits, xtifficon_width, xtifficon_height));
    XtSetArg(args[4], XtNallowShellResize, True);
    XtSetValues(shellWidget, args, 5);

    /*
     * widget instance hierarchy
     */
    formWidget = XtCreateManagedWidget("form", formWidgetClass,
        shellWidget, formArgs, XtNumber(formArgs));

	/*
      	  the box, label, image and pop-up shell belong to the form
	*/
	  widget_list[0] = boxWidget = XtCreateWidget("box",
	    boxWidgetClass, formWidget, boxArgs, XtNumber(boxArgs));

          widget_list[1] = labelWidget = XtCreateWidget("label",
            labelWidgetClass, formWidget, labelArgs, XtNumber(labelArgs));

          widget_list[2] = imageWidget = XtCreateWidget("image",
            widgetClass, formWidget, imageArgs, XtNumber(imageArgs));

	  widget_list[3] = popupWidget = XtCreatePopupShell("Select Page",
	    transientShellWidgetClass, formWidget, popupArgs, 
	    XtNumber(popupArgs));


           XtManageChildren(widget_list, 4);

	/*
	  the dialogue belongs to the pop-up shell
	*/
	  dialogueWidget = XtCreateWidget("dialogue", dialogWidgetClass, 
	    popupWidget, dialogueArgs, XtNumber(dialogueArgs));

	  XtManageChild(dialogueWidget);

	/*
	  the command buttons belong in the box
	*/
          for(w=0; w<15; w++) {

		widget_list[w] = XtCreateWidget("button", commandWidgetClass,
					boxWidget, NULL, 0);

		buttonpixmaps[w] = XCreateBitmapFromData(xDisplay, 
					         RootWindow(xDisplay, xScreen),
					         button_bitmaps[w], 
					         button_width, button_height);

		XtSetArg(args[0], XtNbitmap, buttonpixmaps[w]);
		XtSetValues(widget_list[w], args, 1);

		XtAddCallback(widget_list[w], XtNcallback, SelectProc, 
				(void *) w );

	  } /* for w */

          XtManageChildren(widget_list, 15);

    /*
     * initial widget sizes - for small images let xtiff size itself
     */
    if (tfImageWidth >= appData.viewportWidth) {
        XtSetArg(args[0], XtNwidth, appData.viewportWidth);
        XtSetValues(shellWidget, args, 1);
    }
    if (tfImageHeight >= appData.viewportHeight) {
        XtSetArg(args[0], XtNheight, appData.viewportHeight);
        XtSetValues(shellWidget, args, 1);
    }

    XtSetArg(args[0], XtNwidth, tfImageWidth);
    XtSetArg(args[1], XtNheight, tfImageHeight);
    XtSetValues(imageWidget, args, 2);

    /*
     * formWidget uses these constraints but they are stored in the children.
     */
    XtSetArg(args[0], XtNfromVert, boxWidget);
    XtSetValues(labelWidget, args, 1);
    XtSetArg(args[0], XtNfromVert, labelWidget);
    XtSetValues(imageWidget, args, 1);

    /*
      buttons on the dialogue box
    */
      XawDialogAddButton(dialogueWidget, "Goto", SelectProc, 
			  (void *) ButtonGotoPage);
      XawDialogAddButton(dialogueWidget, "Cancel", SelectProc, 
			  (void *) ButtonDontGotoPage);

    if( -1 == appData.zf  )
	GetBestFit();
    else
	setzoomfactor(appData.zf);

    SetNameLabel();

    /*
     * This is intended to be a little faster than going through
     * the translation manager.
     */
    XtAddEventHandler(imageWidget, ExposureMask | ButtonPressMask
        | ButtonReleaseMask | Button1MotionMask | KeyPressMask,
        False, EventProc, NULL);

    XtRealizeWidget(shellWidget);
    XtRealizeWidget(popupWidget);

    busy(True);
    CreateXImage(INITSTATE);
    busy(False);

    XtMainLoop();
}

void
busy(be_busy)
Boolean	be_busy;
{
 XSetWindowAttributes	wa;

 if( True == be_busy ) {

 	wa.cursor = XCreateFontCursor(xDisplay, XC_watch);
	XChangeWindowAttributes(xDisplay, XtWindow(boxWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(imageWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(labelWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(formWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(popupWidget),
		CWCursor, &wa);

 } /* if True */
 else {

 	wa.cursor = XCreateFontCursor(xDisplay, XC_fleur);
	XChangeWindowAttributes(xDisplay, XtWindow(imageWidget),
		CWCursor, &wa);
	wa.cursor = None;
	XChangeWindowAttributes(xDisplay, XtWindow(boxWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(labelWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(formWidget),
		CWCursor, &wa);
	XChangeWindowAttributes(xDisplay, XtWindow(popupWidget),
		CWCursor, &wa);

 } /* if True ... else */

 XSync(xDisplay, False);

} /* busy */

void
OpenTIFFFile()
{
    if (tfFile != NULL)
        TIFFClose(tfFile);

    if ((tfFile = TIFFOpen(fileName, "r")) == NULL) {
        if (appData.verbose)
            (void)fprintf(stderr, "xtiff: can't open %s as a TIFF file\n", 
								fileName);
        else
            (void)fprintf(stdout, "xtiff: can't open %s as a TIFF file\n",
								 fileName);
        exit(0);
    }

}

void
CountTIFFPages()
{
 int	page = 1;

 while( True == TIFFReadDirectory(tfFile) )
	++page;

 TIFFSetDirectory(tfFile, tfDirectory = 0);

 tfNumPages = page;

} /* CountTIFFPages */


void
GetTIFFHeader()
{
    register int i;

    if (True != TIFFSetDirectory(tfFile, tfDirectory)) {
        (void)fprintf(stderr, "xtiff: can't seek to directory %d in %s\n",
            tfDirectory, fileName);
        exit(0);
    }

    TIFFGetField(tfFile, TIFFTAG_IMAGEWIDTH, &tfImageWidth);
    TIFFGetField(tfFile, TIFFTAG_IMAGELENGTH, &tfImageHeight);

    /*
     * If the following tags aren't present then use the TIFF defaults.
     */
    if (!TIFFGetField(tfFile, TIFFTAG_BITSPERSAMPLE, &tfBitsPerSample))
        tfBitsPerSample = 1;
    if (!TIFFGetField(tfFile, TIFFTAG_SAMPLESPERPIXEL, &tfSamplesPerPixel))
        tfSamplesPerPixel = 1;
    if (!TIFFGetField(tfFile, TIFFTAG_PLANARCONFIG, &tfPlanarConfiguration))
        tfPlanarConfiguration = PLANARCONFIG_CONTIG;
    if (!TIFFGetField(tfFile, TIFFTAG_GRAYRESPONSEUNIT, &tfGrayResponseUnit))
        tfGrayResponseUnit = 2;

    tfUnitMap = tfGrayResponseUnitMap[tfGrayResponseUnit];
    colormapSize = 1 << tfBitsPerSample;
    tfImageDepth = tfBitsPerSample * tfSamplesPerPixel;

    dRed = (double *) malloc(colormapSize * sizeof(double));
    dGreen = (double *) malloc(colormapSize * sizeof(double));
    dBlue = (double *) malloc(colormapSize * sizeof(double));
    MCHECK(dRed); MCHECK(dGreen); MCHECK(dBlue);

    /*
     * If TIFFTAG_PHOTOMETRIC is not present then assign a reasonable default.
     * The TIFF 5.0 specification doesn't give a default.
     */
    if (!TIFFGetField(tfFile, TIFFTAG_PHOTOMETRIC,
            &tfPhotometricInterpretation)) {
        if (tfSamplesPerPixel != 1)
            tfPhotometricInterpretation = PHOTOMETRIC_RGB;
        else if (tfBitsPerSample == 1)
            tfPhotometricInterpretation = PHOTOMETRIC_MINISBLACK;
        else if (TIFFGetField(tfFile, TIFFTAG_COLORMAP,
                &redMap, &greenMap, &blueMap)) {
            tfPhotometricInterpretation = PHOTOMETRIC_PALETTE;
            redMap = greenMap = blueMap = NULL;
        } else
            tfPhotometricInterpretation = PHOTOMETRIC_MINISBLACK;
    }

    /*
     * Given TIFFTAG_PHOTOMETRIC extract or create the response curves.
     */
    switch (tfPhotometricInterpretation) {
    case PHOTOMETRIC_RGB:
        if (!TIFFGetField(tfFile, TIFFTAG_COLORRESPONSECURVE,
                &redMap, &greenMap, &blueMap)) {
            redMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            greenMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            blueMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            MCHECK(redMap); MCHECK(greenMap); MCHECK(blueMap);
            for (i = 0; i < colormapSize; i++)
                dRed[i] = dGreen[i] = dBlue[i]
                    = (double) SCALE(i, colormapSize - 1);
        } else {
            CheckAndCorrectColormap();
            for (i = 0; i < colormapSize; i++) {
                dRed[i] = (double) redMap[i];
                dGreen[i] = (double) greenMap[i];
                dBlue[i] = (double) blueMap[i];
            }
        }
        break;
    case PHOTOMETRIC_PALETTE:
        if (!TIFFGetField(tfFile, TIFFTAG_COLORMAP,
                &redMap, &greenMap, &blueMap)) {
            redMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            greenMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            blueMap = (u_short *) malloc(colormapSize * sizeof(u_short));
            MCHECK(redMap); MCHECK(greenMap); MCHECK(blueMap);
            for (i = 0; i < colormapSize; i++)
                dRed[i] = dGreen[i] = dBlue[i]
                    = (double) SCALE(i, colormapSize - 1);
        } else {
            CheckAndCorrectColormap();
            for (i = 0; i < colormapSize; i++) {
                dRed[i] = (double) redMap[i];
                dGreen[i] = (double) greenMap[i];
                dBlue[i] = (double) blueMap[i];
            }
        }
        break;
    case PHOTOMETRIC_MINISWHITE:
        redMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        greenMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        blueMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        MCHECK(redMap); MCHECK(greenMap); MCHECK(blueMap);
        if (!TIFFGetField(tfFile, TIFFTAG_GRAYRESPONSECURVE, &grayMap))
            for (i = 0; i < colormapSize; i++)
                dRed[i] = dGreen[i] = dBlue[i]
                    = (double) SCALE(colormapSize - 1 - i, colormapSize - 1);
        else {
            dRed[colormapSize - 1] = dGreen[colormapSize - 1]
                = dBlue[colormapSize - 1] = 0;
            for (i = 0; i < colormapSize - 1; i++)
                dRed[i] = dGreen[i] = dBlue[i] = 65535.0 -
                    (65535.0 * pow(10.0, (double) grayMap[i] / tfUnitMap));
        }
        break;
    case PHOTOMETRIC_MINISBLACK:
        redMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        greenMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        blueMap = (u_short *) malloc(colormapSize * sizeof(u_short));
        MCHECK(redMap); MCHECK(greenMap); MCHECK(blueMap);
        if (!TIFFGetField(tfFile, TIFFTAG_GRAYRESPONSECURVE, &grayMap))
            for (i = 0; i < colormapSize; i++) {
                dRed[i] = dGreen[i] = dBlue[i]
                    = (double) SCALE(i, colormapSize - 1);
            }
        else {
            dRed[0] = dGreen[0] = dBlue[0] = 0;
            for (i = 1; i < colormapSize; i++)
                dRed[i] = dGreen[i] = dBlue[i] =
                    (65535.0 * pow(10.0, (double) grayMap[i] / tfUnitMap));
        }
        break;
    default:
        (void)fprintf(stderr,
            "xtiff: can't display photometric interpretation type %d\n",
            tfPhotometricInterpretation);
        exit(0);
    }
}

void
SetNameLabel()
{
    char buffer[BUFSIZ];
    Arg args[1];

    if ( 1 < tfNumPages )
	(void)sprintf(buffer, 
			"%s - page %d of %d - 1/%d size - rotated %d degrees",
 					      fileName, 
					      tfDirectory + 1,
					      tfNumPages,
					      getzoomfactor(),
					      getrotationfactor());
    else
        (void)sprintf(buffer, "%s - 1/%d size - rotated %d degrees",
					      fileName, 
					      getzoomfactor(),
					      getrotationfactor());
    XtSetArg(args[0], XtNlabel, buffer);
    XtSetValues(labelWidget, args, 1);
}

/*
 * Many programs get TIFF colormaps wrong.  They use 8-bit colormaps instead of
 * 16-bit colormaps.  This function is a heuristic to detect and correct this.
 */
void
CheckAndCorrectColormap()
{
    register int i;

    for (i = 0; i < colormapSize; i++)
        if ((redMap[i] > 255) || (greenMap[i] > 255) || (blueMap[i] > 255))
            return;

    for (i = 0; i < colormapSize; i++) {
        redMap[i] = SCALE(redMap[i], 255);
        greenMap[i] = SCALE(greenMap[i], 255);
        blueMap[i] = SCALE(blueMap[i], 255);
    }
    TIFFWarning(fileName, "Assuming 8-bit colormap");
}

void
SimpleGammaCorrection()
{
    register int i;
    register double i_gamma = 1.0 / appData.gamma;

    for (i = 0; i < colormapSize; i++) {
        if (((tfPhotometricInterpretation == PHOTOMETRIC_MINISWHITE)
            && (i == colormapSize - 1))
            || ((tfPhotometricInterpretation == PHOTOMETRIC_MINISBLACK)
            && (i == 0)))
            redMap[i] = greenMap[i] = blueMap[i] = 0;
        else {
            redMap[i] = ROUND((pow(dRed[i] / 65535.0, i_gamma) * 65535.0));
            greenMap[i] = ROUND((pow(dGreen[i] / 65535.0, i_gamma) * 65535.0));
            blueMap[i] = ROUND((pow(dBlue[i] / 65535.0, i_gamma) * 65535.0));
        }
    }

    free(dRed); free(dGreen); free(dBlue);
}

static char* classNames[] = {
    "StaticGray",
    "GrayScale",
    "StaticColor",
    "PseudoColor",
    "TrueColor",
    "DirectColor"
};

/*
 * Current limitation: the visual is set initially by the first file.
 * It cannot be changed.
 */
void
GetVisual()
{
    register XColor *colors = NULL;
    register u_long *pixels = NULL;
    register int i;

    switch (tfImageDepth) {
    /*
     * X really wants a 32-bit image with the fourth channel unused,
     * but the visual structure thinks it's 24-bit.  bitmap_unit is 32.
     */
    case 32:
    case 24:
        if (SearchVisualList(24, DirectColor, &xVisual) == False) {
            (void)fprintf(stderr, 
			"xtiff: 24-bit DirectColor visual not available\n");
            exit(0);
        }

        colors = (XColor *) malloc(3 * colormapSize * sizeof(XColor));
        MCHECK(colors);

        for (i = 0; i < colormapSize; i++) {
            colors[i].pixel = (u_long) (i << 16) + (i << 8) + i;
            colors[i].red = redMap[i];
            colors[i].green = greenMap[i];
            colors[i].blue = blueMap[i];
            colors[i].flags = DoRed | DoGreen | DoBlue;
        }

        xColormap = XCreateColormap(xDisplay, RootWindow(xDisplay, xScreen),
            xVisual, AllocAll);
        XStoreColors(xDisplay, xColormap, colors, colormapSize);
        break;
    case 8:
    case 4:
    case 2:
        /*
         * We assume that systems with 24-bit visuals also have 8-bit visuals.
         * We don't promote from 8-bit PseudoColor to 24/32 bit DirectColor.
         */
        switch (tfPhotometricInterpretation) {
        case PHOTOMETRIC_MINISWHITE:
        case PHOTOMETRIC_MINISBLACK:
            if (SearchVisualList((int) tfImageDepth, GrayScale, &xVisual) == True)
                break;
        case PHOTOMETRIC_PALETTE:
            if (SearchVisualList((int) tfImageDepth, PseudoColor, &xVisual) == True)
                break;
        default:
            (void)fprintf(stderr, "xtiff: Unsupported TIFF/X configuration\n");
            exit(0);
        }

        colors = (XColor *) malloc(colormapSize * sizeof(XColor));
        MCHECK(colors);

        for (i = 0; i < colormapSize; i++) {
            colors[i].pixel = (u_long) i;
            colors[i].red = redMap[i];
            colors[i].green = greenMap[i];
            colors[i].blue = blueMap[i];
            colors[i].flags = DoRed | DoGreen | DoBlue;
        }

        /*
         * xtiff's colormap allocation is private.  It does not attempt
         * to detect whether any existing colormap entries are suitable
         * for its use.  This will cause colormap flashing.  Furthermore,
         * background and foreground are taken from the environment.
         * For example, the foreground color may be red when the visual
         * is GrayScale.  If the colormap is completely populated,
         * Xt will not be able to allocate fg and bg.
         */
        if (tfImageDepth == 8)
            xColormap = XCreateColormap(xDisplay, RootWindow(xDisplay, xScreen),
                xVisual, AllocAll);
        else {
            xColormap = XCreateColormap(xDisplay, RootWindow(xDisplay, xScreen),
                xVisual, AllocNone);
            pixels = (u_long *) malloc(colormapSize * sizeof(u_long));
            MCHECK(pixels);
            (void) XAllocColorCells(xDisplay, xColormap, True,
                NULL, 0, pixels, colormapSize);
            basePixel = (u_char) pixels[0];
            free(pixels);
        }
        XStoreColors(xDisplay, xColormap, colors, colormapSize);
        break;
    case 1:
        xImageDepth = 1;
        xVisual = DefaultVisual(xDisplay, xScreen);
        xColormap = DefaultColormap(xDisplay, xScreen);
        break;
    default:
        (void)fprintf(stderr, "xtiff: unsupported image depth %d\n", 
								tfImageDepth);
        exit(0);
    }

    if (appData.verbose == True)
	(void)fprintf(stderr, "%s: Using %d-bit %s visual.\n",
	    fileName, xImageDepth, classNames[xVisual->class]);

    if (colors != NULL)
        free(colors);
    if (grayMap != NULL)
        free(grayMap);
    if (redMap != NULL)
        free(redMap);
    if (greenMap != NULL)
        free(greenMap);
    if (blueMap != NULL)
        free(blueMap);

    colors = NULL; grayMap = redMap = greenMap = blueMap = NULL;
}

/*
 * Search for an appropriate visual.  Promote where necessary.
 * Check to make sure that ENOUGH colormap entries are writeable.
 * basePixel was determined when XAllocColorCells() contiguously
 * allocated enough entries.  basePixel is used below in GetTIFFImage.
 */
Boolean
SearchVisualList(image_depth, visual_class, visual)
    int image_depth, visual_class;
    Visual **visual;
{
    XVisualInfo template_visual, *visual_list, *vl;
    int i, n_visuals;

    template_visual.screen = xScreen;
    vl = visual_list = XGetVisualInfo(xDisplay, VisualScreenMask,
        &template_visual, &n_visuals);

    if (n_visuals == 0) {
        (void)fprintf(stderr, "xtiff: visual list not available\n");
        exit(0);
    }

    for (i = 0; i < n_visuals; vl++, i++) {
        if ((vl->class == visual_class) && (vl->depth >= image_depth)
            && (vl->visual->map_entries >= (1 << vl->depth))) {
            *visual = vl->visual;
            xImageDepth = vl->depth;
            xRedMask = vl->red_mask;
            xGreenMask = vl->green_mask;
            xBlueMask = vl->blue_mask;
            XFree((char *) visual_list);
            return True;
        }
    }

    XFree((char *) visual_list);
    return False;
}

/*
  allocate enough memory to hold the image at 0, 180 degree rotations and
  90, 270 degree rotations.  if the image doesn't fit neatly into a number of
  bytes there'll be some padding at the end. the padding will only be one
  byte (containing the spare bits) on the end of a scan line, but when rotated 
  will be as many scan lines as there are spare bits
*/  
u_long
allocimage()
{
 u_long		bytes,
		xbytes,
		ybytes;

 if( !(imagebitsize % 8) ) 
	bytes = tfImageWidth * tfImageHeight * (imagebitsize / 8);
 else {
	xbytes =   (tfImageWidth * imagebitsize) / 8  
	       		    + ((tfImageWidth * imagebitsize) % 8 ? 1 : 0);
	ybytes =   (tfImageHeight * imagebitsize) / 8  
	       		    + ((tfImageHeight * imagebitsize) % 8 ? 1 : 0);

	seteffectivewidth(xbytes * 8);
	seteffectiveheight(ybytes * 8);

	bytes = MAX( xbytes * tfImageHeight, ybytes * tfImageWidth );

 } /* if imagebitsize ... else */

 imageMemory = (u_char *) malloc(bytes);
 MCHECK(imageMemory);

 return bytes;

} /* allocimage */

void
GetTIFFImage()
{
    int pixel_map[3], red_shift, green_shift, blue_shift;
    register u_char *scan_line, *output_p, *input_p;
    register int i, j, s;

    if( NULL != imageMemory ) {
	free(imageMemory);
	imageMemory = NULL;
    } /* if NULL */

    scan_line = (u_char *) malloc((tfBytesPerRow = TIFFScanlineSize(tfFile)));
    MCHECK(scan_line);

    if ((tfImageDepth == 32) || (tfImageDepth == 24)) {

	imagebitsize = 32;

	isize = allocimage();
        output_p = imageMemory;

        /*
         * Handle different color masks for different frame buffers.
         */
        if (ImageByteOrder(xDisplay) == LSBFirst) { /* DECstation 5000 */
            red_shift = pixel_map[0] = xRedMask == 0xFF000000 ? 3
                : (xRedMask == 0xFF0000 ? 2 : (xRedMask == 0xFF00 ? 1 : 0));
            green_shift = pixel_map[1] = xGreenMask == 0xFF000000 ? 3
                : (xGreenMask == 0xFF0000 ? 2 : (xGreenMask == 0xFF00 ? 1 : 0));
            blue_shift = pixel_map[2] = xBlueMask == 0xFF000000 ? 3
                : (xBlueMask == 0xFF0000 ? 2 : (xBlueMask == 0xFF00 ? 1 : 0));
        } else { /* Ardent */
            red_shift = pixel_map[0] = xRedMask == 0xFF000000 ? 0
                : (xRedMask == 0xFF0000 ? 1 : (xRedMask == 0xFF00 ? 2 : 3));
            green_shift = pixel_map[0] = xGreenMask == 0xFF000000 ? 0
                : (xGreenMask == 0xFF0000 ? 1 : (xGreenMask == 0xFF00 ? 2 : 3));
            blue_shift = pixel_map[0] = xBlueMask == 0xFF000000 ? 0
                : (xBlueMask == 0xFF0000 ? 1 : (xBlueMask == 0xFF00 ? 2 : 3));
        }

        if (tfPlanarConfiguration == PLANARCONFIG_CONTIG) {

            for (i = 0; i < tfImageHeight; i++) {

                TIFFReadScanline(tfFile, scan_line, i, 0);

                for (input_p = scan_line, j = 0; j < tfImageWidth; j++) {
                    *(output_p + red_shift) = *input_p++;
                    *(output_p + green_shift) = *input_p++;
                    *(output_p + blue_shift) = *input_p++;
                    output_p += 4;
                    if (tfSamplesPerPixel == 4) /* skip the fourth channel */
                        input_p++;
                }
            }
        } else {

            for (s = 0; s < tfSamplesPerPixel; s++) {
                if (s == 3)             /* skip the fourth channel */
                    continue;
                for (i = 0; i < tfImageHeight; i++) {

                    TIFFReadScanline(tfFile, scan_line, i, s);

                    input_p = scan_line;
                    output_p = imageMemory + (i*tfImageWidth*4) + pixel_map[s];
                    for (j = 0; j < tfImageWidth; j++, output_p += 4)
                        *output_p = *input_p++;
                }
            }
        }
    } else {
        if (xImageDepth == tfImageDepth) {

	    imagebitsize = xImageDepth;

	    isize = allocimage();
            output_p = imageMemory;

            for (i = 0; i < tfImageHeight; i++, output_p += tfBytesPerRow) 
                TIFFReadScanline(tfFile, output_p, i, 0); 

        } else if ((xImageDepth == 8) && (tfImageDepth == 4)) {

	    imagebitsize = 8;

	    isize = allocimage();
            output_p = imageMemory;

            /*
             * If a scanline is of odd size the inner loop below will overshoot.
             * This is handled very simply by recalculating the start point at
             * each scanline and padding imageMemory a little at the end.
             */
            for (i = 0; i < tfImageHeight; i++) {
                TIFFReadScanline(tfFile, scan_line, i, 0);

                output_p = &imageMemory[i * tfImageWidth];
                input_p = scan_line;

                for (j = 0; j < tfImageWidth; j += 2, input_p++) {
                    *output_p++ = (*input_p >> 4) + basePixel;
                    *output_p++ = (*input_p & 0xf) + basePixel;
                }
            }
        } else if ((xImageDepth == 8) && (tfImageDepth == 2)) {

	    imagebitsize = 8;

	    isize = allocimage();
            output_p = imageMemory;

            for (i = 0; i < tfImageHeight; i++) {
                TIFFReadScanline(tfFile, scan_line, i, 0);

                output_p = &imageMemory[i * tfImageWidth];
                input_p = scan_line;
                for (j = 0; j < tfImageWidth; j += 4, input_p++) {
                    *output_p++ = (*input_p >> 6) + basePixel;
                    *output_p++ = ((*input_p >> 4) & 3) + basePixel;
                    *output_p++ = ((*input_p >> 2) & 3) + basePixel;
                    *output_p++ = (*input_p & 3) + basePixel;
                }
            }
        } else if ((xImageDepth == 4) && (tfImageDepth == 2)) {

	    imagebitsize = 4;

	    isize = allocimage();
            output_p = imageMemory;

            for (i = 0; i < tfImageHeight; i++) {
                TIFFReadScanline(tfFile, scan_line, i, 0);

                output_p = &imageMemory[i * tfBytesPerRow * 2];
                input_p = scan_line;

                for (j = 0; j < tfImageWidth; j += 4, input_p++) {
                    *output_p++ = (((*input_p>>6) << 4)
                        | ((*input_p >> 4) & 3)) + basePixel;
                    *output_p++ = ((((*input_p>>2) & 3) << 4)
                        | (*input_p & 3)) + basePixel;
                }
            }
        } else {
            (void)fprintf(stderr,
                "xtiff: can't handle %d-bit TIFF file on an %d-bit display\n",
                tfImageDepth, xImageDepth);
            exit(0);
        }
    }

    free(scan_line);

}

void
CreateXImage(state)
int state;
{
    XGCValues 	gc_values;
    GC 		bitmap_gc;


    xOffset = yOffset = 0;
    grabX = grabY = -1;

    if( !transformimage(state) ) 
	return;

    xImage = XCreateImage(xDisplay, xVisual, xImageDepth,
		xImageDepth == 1 ? XYBitmap : ZPixmap, /* offset */ 0,
		(char *) zoomMemory, getwidth(), getheight(), 
		/* bitmap_pad */ 8, 
                /* bytes_per_line */  0 );

    /*
     * libtiff converts LSB data into MSB but doesn't change the FillOrder tag.
     */
    if (xImageDepth == 1)
        xImage->bitmap_bit_order = MSBFirst;
    if (xImageDepth <= 8)
        xImage->byte_order = MSBFirst;

    /*
     * create an appropriate GC
     */
    gc_values.function = GXcopy;
    gc_values.plane_mask = AllPlanes;
    if (tfPhotometricInterpretation == PHOTOMETRIC_MINISBLACK) {
        gc_values.foreground = XWhitePixel(xDisplay, xScreen);
        gc_values.background = XBlackPixel(xDisplay, xScreen);
    } else {
        gc_values.foreground = XBlackPixel(xDisplay, xScreen);
        gc_values.background = XWhitePixel(xDisplay, xScreen);
    }
    xWinGc = XCreateGC(xDisplay, XtWindow(shellWidget),
        GCFunction | GCPlaneMask | GCForeground | GCBackground, &gc_values);

    /*
     * create the pixmap and load the image
     */
    if (appData.usePixmap == True) {
        xImagePixmap = XCreatePixmap(xDisplay, RootWindow(xDisplay, xScreen),
            xImage->width, xImage->height, xImageDepth);

        /*
         * According to the O'Reilly X Protocol Reference Manual, page 53,
         * "A pixmap depth of one is always supported and listed, but windows
         * of depth one might not be supported."  Therefore we create a pixmap
         * of depth one and use XCopyPlane().  This is idiomatic.
         */
        if (xImageDepth == 1) {         /* just pass the bits through */
            gc_values.foreground = 1;   /* foreground describes set bits */
            gc_values.background = 0;   /* background describes clear bits */
            bitmap_gc = XCreateGC(xDisplay, xImagePixmap,
                GCForeground | GCBackground, &gc_values);
            XPutImage(xDisplay, xImagePixmap, bitmap_gc, xImage,
                0, 0, 0, 0, xImage->width, xImage->height);
        } else
            XPutImage(xDisplay, xImagePixmap, xWinGc, xImage,
                0, 0, 0, 0, xImage->width, xImage->height);

        XDestroyImage(xImage);
    }
}

/* ARGSUSED */
XtCallbackProc
SelectProc(w, button_id, unused_2)
    Widget w;
    caddr_t button_id;
    caddr_t unused_2;
{
    Arg		args[1];
    char	pagestring[256];

    busy(True);

    switch ((int) button_id) {
    case ButtonQuit:
	busy(False);
        QuitProc();
        break;
    case ButtonPreviousPage:
        PreviousProc();
        break;
    case ButtonNextPage:
        NextProc();
        break;
    case ButtonGrow:
	GrowProc();
	break;
    case ButtonShrink:
	ShrinkProc();
	break;
    case ButtonLargest:
	LargestProc();
	break;
    case ButtonBestFit:
	BestFitProc();
	break;	
    case Button0:
	ZeroProc();
	break;
    case Button90:
	NinetyProc();
	break;
    case Button180:
	OneEightyProc();
	break;
    case Button270:
	TwoSeventyProc();
	break;
    case ButtonFirstPage:
	FirstPageProc();
	break;
    case ButtonLastPage:
	LastPageProc();
	break;
    case ButtonGotoDialoguePopup:
	(void)sprintf(pagestring, "%d", tfDirectory + 1);
	XtSetArg(args[0], XtNvalue, pagestring);
	XtSetValues(dialogueWidget, args, 1);
	XtPopup(popupWidget, XtGrabExclusive);
	break;
    case ButtonHelp:
	behelpful();
	break;
    case ButtonGotoPage:
	busy(True);
	XtPopdown(popupWidget);
	GotoPageProc();
	break;
    case ButtonDontGotoPage:
	XtPopdown(popupWidget);
	busy(False);
	return;
    default:
        (void)fprintf(stderr, "xtiff: unknown button %d in SelectProc\n", 
			(int) button_id);
	exit(1);
    }
    busy(False);
}

void
QuitProc()
{
    exit(0);
}

void
NextProc()
{
    PageProc(ButtonNextPage);
}

void
PreviousProc()
{
    PageProc(ButtonPreviousPage);
}

void 
ShrinkProc()
{
 zoomproc(ZOOMDELTA, 1);
}

void
GrowProc()
{
 zoomproc(ZOOMDELTA, -1);
} 

void
LargestProc()
{
 zoomproc(ZOOMSET, 1);
} 


/*
  this could be improved - most of the CERN preprint scans are surrounded by
  large, white borders, which hardly need to fit into the frame
*/
void
GetBestFit()
{
 Arg		args[2];
 int 		nz = 1,
		ww,
		wh;
 Dimension  	w_height,
		w_width;

 XtSetArg(args[0], XtNwidth, &w_width);
 XtSetArg(args[1], XtNheight, &w_height);
 XtGetValues(shellWidget, args, 2);
 ww = w_width;
 wh = w_height;
 XtGetValues(boxWidget, args, 2);
 wh -= w_height;
 XtGetValues(labelWidget, args, 2);
 wh -= w_height;

 while( getrealwidth() / nz > ww || getrealheight() / nz > wh )
	++nz;

 setzoomfactor(nz);
} /* GetBestFit */


void
BestFitProc()
{ 
 GetBestFit();
 
 redisplay(ZOOMCHANGE);
} /* BestFitProc */

void
ZeroProc()
{
 rotationproc(0);
} /* ZeroProc */

void
NinetyProc()
{
 rotationproc(90);
} /* NinetyProc */

void
OneEightyProc()
{
 rotationproc(180);
} /* OneEightyProc */

void
TwoSeventyProc()
{
 rotationproc(270);
} /* TwoSeventyProc */ 

void
FirstPageProc()
{
 PageProc(ButtonFirstPage);
} /* FirstPageProc */

void 
LastPageProc()
{
 PageProc(ButtonLastPage);
} /* LastPageProc */

void
GotoPageProc()
{
 PageProc(ButtonGotoPage);
} /* GotoPageProc */


void
redisplay(status)
int status;
{
    XEvent fake_event;
    Arg args[2];

    SetNameLabel();

    if (appData.usePixmap == True)
	XFreePixmap(xDisplay, xImagePixmap);
    else
	XDestroyImage(xImage);

    CreateXImage(status);

    /*
     * Using XtSetValues() to set the widget size causes a resize.
     * This resize gets propagated up to the parent shell.
     * In order to disable this visually disconcerting effect,
     * shell resizing is temporarily disabled.
     */
    XtSetArg(args[0], XtNallowShellResize, False);
    XtSetValues(shellWidget, args, 1);
    
    XtSetArg(args[0], XtNwidth, getrealwidth());
    XtSetArg(args[1], XtNheight,getrealheight());
    XtSetValues(imageWidget, args, 2);

    XtSetArg(args[0], XtNallowShellResize, True);
    XtSetValues(shellWidget, args, 1);
    
    XClearWindow(xDisplay, XtWindow(imageWidget));

    fake_event.type = Expose;
    fake_event.xexpose.x = fake_event.xexpose.y = 0;
    fake_event.xexpose.width = getrealwidth() ;  /* the window will clip */
    fake_event.xexpose.height = getrealheight();
    EventProc(imageWidget, NULL, &fake_event);

} /* redisplay */

void
zoomproc(flag, zoomordelta)
int flag;
int zoomordelta;
{
    if( ZOOMDELTA == flag ) { 
    	if( 0 == (getzoomfactor() + zoomordelta) )  {
		setzoomfactor(1);
		return;
    	} /* if 0 */

	setzoomfactor( getzoomfactor() + zoomordelta );

    } /* if ZOOMDELTA */
    else 
	setzoomfactor(zoomordelta);

    redisplay(ZOOMCHANGE);

}

void
rotationproc(rotationfactor)
int rotationfactor;
{
    setrotationfactor(rotationfactor);

    redisplay(ROTATECHANGE);

} /* rotateproc */

void
PageProc(direction)
    int direction;
{
    int		newdir;

    Arg 	args[1];
    char	pagestring[256];

    switch (direction) {

    case ButtonFirstPage:
	if( 0 < tfDirectory )
		tfDirectory = 0;	/* GetTIFFHeader() sets dir */
	else 
		return;
	break; 

    case ButtonPreviousPage:
        if (tfDirectory > 0) 		/* GetTIFFHeader() sets dir */
            --tfDirectory;
        else 
            return;
        break;

    case ButtonNextPage:
	if( tfNumPages > tfDirectory + 1 ) 
		++tfDirectory;		/* GetTIFFHeader() sets dir */
	else
		return;
        break;

    case ButtonLastPage:
	if( tfDirectory + 1 != tfNumPages )
		tfDirectory = tfNumPages - 1;	/* GetTIFFHeader() sets dir */
	else
		return;

	break;

    case ButtonGotoPage:
	/* 
	  nothing to do if we can't read a number ...
	*/
	  if(!sscanf(XawDialogGetValueString(dialogueWidget), "%d", &newdir)) { 

	        (void)sprintf(pagestring, "%d", tfDirectory + 1);
	        XtSetArg(args[0], XtNvalue, pagestring);
	        XtSetValues(dialogueWidget, args, 1);

		return;

	  } /* if !sscanf */

	if( 1 > newdir || tfNumPages < newdir || tfDirectory + 1 == newdir )
		return;
	else
		tfDirectory = newdir - 1;	/* GetTIFFHeader() sets dir */

	break;

    default:
        (void)fprintf(stderr, "error in PageProc\n");
        exit(0);
    }

    xOffset = yOffset = 0;
    grabX = grabY = -1;

    GetTIFFHeader();
    GetTIFFImage();

    (void)sprintf(pagestring, "%d", tfDirectory + 1);
    XtSetArg(args[0], XtNvalue, pagestring);
    XtSetValues(dialogueWidget, args, 1);

    redisplay(PAGECHANGE);
}

/* ARGSUSED */
void
EventProc(widget, unused, event)
    Widget widget;
    caddr_t unused;
    XEvent *event;
{
    int ih, iw, ww, wh, sx, sy, w, h, dx, dy;
    Dimension w_width, w_height;
    XEvent next_event;
    Arg args[2];

    if (event->type == MappingNotify) {
        XRefreshKeyboardMapping((XMappingEvent *) event);
        return;
    }

    if (!XtIsRealized(widget))
        return;

    if ((event->type == ButtonPress) || (event->type == ButtonRelease))
        if (event->xbutton.button != Button1)
            return;

    iw = getwidth() ;  /* avoid sign problems */
    ih = getheight() ;

    /*
     * The grabX and grabY variables record where the user grabbed the image.
     * They also record whether the mouse button is down or not.
     */
    if (event->type == ButtonPress) {
        grabX = event->xbutton.x;
        grabY = event->xbutton.y;
        return;
    }

    /*
     * imageWidget is a Core widget and doesn't get resized.
     * So we calculate the size of its viewport here.
     */
    XtSetArg(args[0], XtNwidth, &w_width);
    XtSetArg(args[1], XtNheight, &w_height);
    XtGetValues(shellWidget, args, 2);
    ww = w_width;
    wh = w_height;
    XtGetValues(boxWidget, args, 2);
    wh -= w_height;

    switch (event->type) {
    case Expose:
        dx = event->xexpose.x;
        dy = event->xexpose.y;
        sx = dx + xOffset;
        sy = dy + yOffset;
        w = MIN(event->xexpose.width, iw);
        h = MIN(event->xexpose.height, ih);
        break;
    case KeyPress:
        if ((grabX >= 0) || (grabY >= 0)) /* Mouse button is still down */
            return;
        switch (XLookupKeysym((XKeyEvent *) event, /* KeySyms index */ 0)) {
        case XK_Up:
            if (ih < wh)    /* Don't scroll if the window fits the image. */
                return;
            sy = yOffset + appData.translate;
            sy = MIN(ih - wh, sy);
            if (sy == yOffset)  /* Filter redundant stationary refreshes. */
                return;
            yOffset = sy;
            sx = xOffset;
            dx = dy = 0;
            w = ww; h = wh;
            break;
        case XK_Down:
            if (ih < wh)
                return;
            sy = yOffset - appData.translate;
            sy = MAX(sy, 0);
            if (sy == yOffset)
                return;
            yOffset = sy;
            sx = xOffset;
            dx = dy = 0;
            w = ww; h = wh;
            break;
        case XK_Left:
            if (iw < ww)
                return;
            sx = xOffset + appData.translate;
            sx = MIN(iw - ww, sx);
            if (sx == xOffset)
                return;
            xOffset = sx;
            sy = yOffset;
            dx = dy = 0;
            w = ww; h = wh;
            break;
        case XK_Right:
            if (iw < ww)
                return;
            sx = xOffset - appData.translate;
            sx = MAX(sx, 0);
            if (sx == xOffset)
                return;
            xOffset = sx;
            sy = yOffset;
            dx = dy = 0;
            w = ww; h = wh;
            break;
	case XK_Help:
	    if( False == amhelpful() )
	    	behelpful();
	    else
	        beunhelpful();
	    break;
        default:
            return;
        }
        break;
    case MotionNotify:
        /*
         * MotionEvent compression.  Ignore multiple motion events.
         * Ignore motion events if the mouse button is up.
         */
        if (XPending(xDisplay)) /* Xlib doesn't flush the output buffer */
            if (XtPeekEvent(&next_event))
                if (next_event.type == MotionNotify)
                    return;
        if ((grabX < 0) || (grabY < 0))
            return;
        sx = xOffset + grabX - (int) event->xmotion.x;
        if (sx >= (iw - ww))    /* clamp x motion but allow y motion */
            sx = iw - ww;
        sx = MAX(sx, 0);
        sy = yOffset + grabY - (int) event->xmotion.y;
        if (sy >= (ih - wh)) /* clamp y motion but allow x motion */
            sy = ih - wh;
        sy = MAX(sy, 0);
        if ((sx == xOffset) && (sy == yOffset))
            return;
        dx = dy = 0;
        w = ww; h = wh;
        break;
    case ButtonRelease:
        xOffset = xOffset + grabX - (int) event->xbutton.x;
        xOffset = MIN(iw - ww, xOffset);
        xOffset = MAX(xOffset, 0);
        yOffset = yOffset + grabY - (int) event->xbutton.y;
        yOffset = MIN(ih - wh, yOffset);
        yOffset = MAX(yOffset, 0);
        grabX = grabY = -1;
    default:
        return;
    }

    if (appData.usePixmap == True) {
        if (xImageDepth == 1)
            XCopyPlane(xDisplay, xImagePixmap, XtWindow(widget),
                xWinGc, sx, sy, w, h, dx, dy, 1);
        else
            XCopyArea(xDisplay, xImagePixmap, XtWindow(widget),
                xWinGc, sx, sy, w, h, dx, dy);
    } else
        XPutImage(xDisplay, XtWindow(widget), xWinGc, xImage,
            sx, sy, dx, dy, w, h);
}


int
XTiffErrorHandler(display, error_event)
    Display *display;
    XErrorEvent *error_event;
{
    char message[80];

    /*
     * Some X servers limit the size of pixmaps.
     */
    if ((error_event->error_code == BadAlloc)
            && (error_event->request_code == X_CreatePixmap))
        (void)fprintf(stderr, "xtiff: requested pixmap too big for display\n");
    else {
        XGetErrorText(display, error_event->error_code, message, 80);
        (void)fprintf(stderr, "xtiff: error code %s\n", message);
    }

    exit(0);
}

void
Usage()
{
    (void)fprintf(stderr, "Usage xtiff: [options] tiff-file\n");
    (void)fprintf(stderr, "\tstandard Xt options\n");
    (void)fprintf(stderr, "\t[-help]\n");
    (void)fprintf(stderr, "\t[-gamma gamma]\n");
    (void)fprintf(stderr, "\t[-usePixmap (True | False)]\n");
    (void)fprintf(stderr, "\t[-viewportWidth pixels]\n");
    (void)fprintf(stderr, "\t[-viewportHeight pixels]\n");
    (void)fprintf(stderr, "\t[-translate pixels]\n");
    (void)fprintf(stderr, "\t[-verbose (True | False)]\n");
    (void)fprintf(stderr, "\t[-zf zoomfactor (-1 == best fit)]\n");
    exit(0);
}


/*** xtiff.c ***/

